<?php

namespace Drupal\wayfinding\Controller;

use Drupal\Component\Plugin\Exception\InvalidPluginDefinitionException;
use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\Core\Access\AccessResult;
use Drupal\Core\Access\AccessResultInterface;
use Drupal\Core\Ajax\AjaxResponse;
use Drupal\Core\Ajax\HtmlCommand;
use Drupal\Core\Ajax\InvokeCommand;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Render\HtmlResponse;
use Drupal\Core\Render\HtmlResponseAttachmentsProcessor;
use Drupal\Core\StringTranslation\StringTranslationTrait;
use Drupal\digital_signage_device\DeviceInterface;
use Drupal\wayfinding\Query;
use Endroid\QrCode\QrCode;
use Symfony\Component\DependencyInjection\ContainerInterface;
use Symfony\Component\HttpFoundation\RequestStack;

/**
 * Provides the Device controller.
 */
class Device implements ContainerInjectionInterface {

  use StringTranslationTrait;

  /**
   * Tbd.
   *
   * @var \Drupal\wayfinding\Query
   */
  protected $query;

  /**
   * Tbd.
   *
   * @var \Drupal\Core\Render\HtmlResponseAttachmentsProcessor
   */
  protected $attachmentProcessor;

  /**
   * Tbd.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  protected $config;

  /**
   * Tbd.
   *
   * @var \Symfony\Component\HttpFoundation\Request
   */
  protected $request;

  /**
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected $entityTypeManager;

  /**
   * Device constructor.
   *
   * @param \Drupal\wayfinding\Query $query
   *   Tbd.
   * @param \Drupal\Core\Render\HtmlResponseAttachmentsProcessor $attachment_processor
   *   Tbd.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $config_factory
   *   Tbd.
   * @param \Symfony\Component\HttpFoundation\RequestStack $request_stack
   *   Tbd.
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   */
  public function __construct(Query $query, HtmlResponseAttachmentsProcessor $attachment_processor, ConfigFactoryInterface $config_factory, RequestStack $request_stack, EntityTypeManagerInterface $entity_type_manager) {
    $this->query = $query;
    $this->attachmentProcessor = $attachment_processor;
    $this->config = $config_factory->get('wayfinding.settings');
    $this->request = $request_stack->getCurrentRequest();
    $this->entityTypeManager = $entity_type_manager;
  }

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): Device {
    return new static(
      $container->get('wayfinding.query'),
      $container->get('html_response.attachments_processor'),
      $container->get('config.factory'),
      $container->get('request_stack'),
      $container->get('entity_type.manager')
    );
  }

  /**
   * Tbd.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Tbd.
   */
  public function access(): AccessResultInterface {
    return AccessResult::allowed();
  }

  /**
   * Tbd.
   *
   * @noinspection PhpUnused
   */

  /**
   * Tbd.
   *
   * @param \Drupal\digital_signage_device\DeviceInterface $digital_signage_device
   *   Tbd.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Tbd.
   */
  public function accessDevice(DeviceInterface $digital_signage_device): AccessResultInterface {
    return $digital_signage_device->isEnabled() ?
      AccessResult::allowed() :
      AccessResult::forbidden();
  }

  /**
   * Tbd.
   *
   * @param string $id
   *   Tbd.
   *
   * @return \Drupal\Core\Access\AccessResultInterface
   *   Tbd.
   */
  public function accessExtDevice($id): AccessResultInterface {
    try {
      $devices = $this->entityTypeManager->getStorage('digital_signage_device')
        ->loadByProperties([
          'extid' => $id,
        ]);
      if (empty($devices)) {
        return AccessResult::forbidden('Unknown external ID');
      }
      if (count($devices) > 1) {
        return AccessResult::forbidden('Ambiguous external ID');
      }
    }
    catch (InvalidPluginDefinitionException $e) {
      return AccessResult::forbidden();
    }
    catch (PluginNotFoundException $e) {
      return AccessResult::forbidden();
    }
    $digital_signage_device = reset($devices);
    return $digital_signage_device->isEnabled() ?
      AccessResult::allowed() :
      AccessResult::forbidden();
  }

  /**
   * Tbd.
   *
   * @return \Drupal\Core\Render\HtmlResponse
   *   Tbd.
   */
  public function request(): HtmlResponse {
    return $this->deliver(0);
  }

  /**
   * Tbd.
   *
   * @noinspection PhpUnused
   */

  /**
   * Tbd.
   *
   * @param \Drupal\digital_signage_device\DeviceInterface $digital_signage_device
   *   Tbd.
   *
   * @return \Drupal\Core\Render\HtmlResponse
   *   Tbd.
   */
  public function requestDevice(DeviceInterface $digital_signage_device): HtmlResponse {
    $position = $digital_signage_device->get('geolocation')->getValue()[0];
    return $this->deliver(
      $digital_signage_device->get('perspective')->value,
      $position['lat'],
      $position['lng'],
      $digital_signage_device
    );
  }

  /**
   * Tbd.
   *
   * @param string $id
   *   Tbd.
   *
   * @return \Drupal\Core\Render\HtmlResponse
   *   Tbd.
   */
  public function requestExtDevice($id): HtmlResponse {
    try {
      $devices = $this->entityTypeManager->getStorage('digital_signage_device')
        ->loadByProperties([
          'extid' => $id,
        ]);
    }
    catch (InvalidPluginDefinitionException $e) {
      // We can ignore this because it has been validated in access callback.
    }
    catch (PluginNotFoundException $e) {
      // We can ignore this because it has been validated in access callback.
    }
    $digital_signage_device = reset($devices);
    return $this->requestDevice($digital_signage_device);
  }

  /**
   * Tbd.
   *
   * @return \Drupal\Core\Ajax\AjaxResponse
   *   Tbd.
   */
  public function widgets(): AjaxResponse {
    // TODO: Sanitize data before processing.
    $data = json_decode($this->request->query->get('data'), TRUE);
    $response = new AjaxResponse();

    if ($this->config->get('enable qr code')) {
      // Build route widgets.
      if (!empty($data['position']['lat']) && !empty($data['position']['lng'])) {
        $routeWidgets = '';
        foreach ($data['destinations'] as $destination) {
          if (!empty($destination['lat']) && !empty($destination['lng'])) {
            $url = 'https://www.google.com/maps/dir/' . $data['position']['lat'] . ',' . $data['position']['lng'] . '/' . $destination['lat'] . ',' . $destination['lng'];
            $qrCode = new QrCode($url);
            $routeWidgets .= '<a href="' . $url . '"><img  alt="Open route in Google Maps" src="data:image/png;base64,' . base64_encode($qrCode->writeString()) . '"></a>';
          }
        }
        if (!empty($routeWidgets)) {
          $response->addCommand(new HtmlCommand('#wayfinding .widgets .widget.route', $routeWidgets));
          $response->addCommand(new InvokeCommand('#wayfinding .widgets .widget.route', 'addClass', ['show']));
        }
      }
    }

    return $response;
  }

  /**
   * Tbd.
   *
   * @param float $perspective
   *   Tbd.
   * @param int $lat
   *   Tbd.
   * @param int $lng
   *   Tbd.
   * @param null $device
   *   Tbd.
   *
   * @return \Drupal\Core\Render\HtmlResponse
   *   Tbd.
   */
  protected function deliver($perspective, $lat = 0, $lng = 0, $device = NULL): HtmlResponse {
    $html_response = new HtmlResponse();
    $html_response->setContent($this->query->build(FALSE, $perspective, $lat, $lng, $device));
    // $html_response->getCacheableMetadata()->setCacheMaxAge(0);
    // $html_response->getCacheableMetadata()->setCacheTags([]);
    // $html_response->getCacheableMetadata()->setCacheContexts([]);
    return $this->attachmentProcessor->processAttachments($html_response);
  }

}
